@@ -2,7 +2,6 @@ package proxy
 
 import (
 	"net"
-	"strings"
 
 	"github.com/AdguardTeam/golibs/log"
 	"github.com/miekg/dns"
@@ -191,18 +190,3 @@ func isPublicIP(ip net.IP) bool {
 
 	return true
 }
-
-// split string by a byte and return the first chunk
-// Whitespace is trimmed
-func splitNext(str *string, splitBy byte) string {
-	i := strings.IndexByte(*str, splitBy)
-	s := ""
-	if i != -1 {
-		s = (*str)[0:i]
-		*str = (*str)[i+1:]
-	} else {
-		s = *str
-		*str = ""
-	}
-	return strings.TrimSpace(s)
-}
